@@ -47,7 +47,7 @@ struct st_engine_ctx {
 	 * must be whitened, to prevent security holes.
 	 */
 	char *pin;
-	int pin_length;
+	size_t pin_length;
 	int verbose;
 	char *module;
 	char *init_args;
@@ -99,7 +99,7 @@ static void dump_hex(ENGINE_CTX *ctx, int level,
 /* Free PIN storage in secure way. */
 static void ctx_destroy_pin(ENGINE_CTX *ctx)
 {
-	if (ctx->pin != NULL) {
+	if (ctx->pin) {
 		OPENSSL_cleanse(ctx->pin, ctx->pin_length);
 		OPENSSL_free(ctx->pin);
 		ctx->pin = NULL;
@@ -118,16 +118,16 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 
 	/* call ui to ask for a pin */
 	ui = UI_new_method(ui_method);
-	if (ui == NULL) {
+	if (!ui) {
 		ctx_log(ctx, 0, "UI_new failed\n");
 		return 0;
 	}
-	if (callback_data != NULL)
+	if (callback_data)
 		UI_add_user_data(ui, callback_data);
 
 	ctx_destroy_pin(ctx);
 	ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
-	if (ctx->pin == NULL)
+	if (!ctx->pin)
 		return 0;
 	memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 	ctx->pin_length = MAX_PIN_LENGTH;
@@ -135,8 +135,8 @@ static int ctx_get_pin(ENGINE_CTX *ctx, const char* token_label, UI_METHOD *ui_m
 	if (!prompt) {
 		return 0;
 	}
-	if (!UI_dup_input_string(ui, prompt,
-			UI_INPUT_FLAG_DEFAULT_PWD, ctx->pin, 4, MAX_PIN_LENGTH)) {
+	if (UI_dup_input_string(ui, prompt,
+			UI_INPUT_FLAG_DEFAULT_PWD, ctx->pin, 4, MAX_PIN_LENGTH) <= 0) {
 		ctx_log(ctx, 0, "UI_dup_input_string failed\n");
 		UI_free(ui);
 		OPENSSL_free(prompt);
@@ -186,7 +186,7 @@ static int ctx_login(ENGINE_CTX *ctx, PKCS11_SLOT *slot, PKCS11_TOKEN *tok,
 		/* Free the PIN if it has already been
 		 * assigned (i.e, cached by ctx_get_pin) */
 		ctx_destroy_pin(ctx);
-	} else if (ctx->pin == NULL) {
+	} else if (!ctx->pin) {
 		ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
 		ctx->pin_length = MAX_PIN_LENGTH;
 		if (ctx->pin == NULL) {
@@ -221,7 +221,7 @@ ENGINE_CTX *ctx_new()
 	char *mod;
 
 	ctx = OPENSSL_malloc(sizeof(ENGINE_CTX));
-	if (ctx == NULL)
+	if (!ctx)
 		return NULL;
 	memset(ctx, 0, sizeof(ENGINE_CTX));
 
@@ -308,7 +308,7 @@ static int ctx_init_libp11(ENGINE_CTX *ctx)
 	if (ctx->rwlock)
 		CRYPTO_w_lock(ctx->rwlock);
 #endif
-	if (ctx->pkcs11_ctx == NULL || ctx->slot_list == NULL)
+	if (!ctx->pkcs11_ctx|| !ctx->slot_list)
 		ctx_init_libp11_unlocked(ctx);
 #if OPENSSL_VERSION_NUMBER >= 0x10100004L && !defined(LIBRESSL_VERSION_NUMBER)
 	CRYPTO_THREAD_unlock(ctx->rwlock);
@@ -375,10 +375,10 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 		const int login)
 {
 	PKCS11_SLOT *slot;
-	PKCS11_SLOT *found_slot = NULL;
+	PKCS11_SLOT *found_slot = NULL, **matched_slots = NULL;
 	PKCS11_TOKEN *tok, *match_tok = NULL;
 	PKCS11_CERT *certs, *selected_cert = NULL;
-	X509 *x509;
+	X509 *x509 = NULL;
 	unsigned int cert_count, n, m;
 	unsigned char cert_id[MAX_VALUE_LEN / 2];
 	size_t cert_id_len = sizeof(cert_id);
@@ -387,6 +387,7 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 	size_t tmp_pin_len = MAX_PIN_LENGTH;
 	int slot_nr = -1;
 	char flags[64];
+	size_t matched_count = 0;
 
 	if (ctx_init_libp11(ctx)) /* Delayed libp11 initialization */
 		return NULL;
@@ -401,14 +402,12 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 					"The certificate ID is not a valid PKCS#11 URI\n"
 					"The PKCS#11 URI format is defined by RFC7512\n");
 				ENGerr(ENG_F_CTX_LOAD_CERT, ENG_R_INVALID_ID);
-				return NULL;
+				goto error;
 			}
 			if (tmp_pin_len > 0 && tmp_pin[0] != 0) {
-				if (!login)
-					return NULL; /* Process on second attempt */
 				ctx_destroy_pin(ctx);
 				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
-				if (ctx->pin != NULL) {
+				if (ctx->pin) {
 					memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 					memcpy(ctx->pin, tmp_pin, tmp_pin_len);
 					ctx->pin_length = tmp_pin_len;
@@ -424,7 +423,7 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 					"The legacy ENGINE_pkcs11 ID format is also "
 					"still accepted for now\n");
 				ENGerr(ENG_F_CTX_LOAD_CERT, ENG_R_INVALID_ID);
-				return NULL;
+				goto error;
 			}
 		}
 		ctx_log(ctx, 1, "Looking in slot %d for certificate: ",
@@ -433,13 +432,20 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 			ctx_log(ctx, 1, "id=");
 			dump_hex(ctx, 1, cert_id, cert_id_len);
 		}
-		if (cert_id_len != 0 && cert_label != NULL)
+		if (cert_id_len != 0 && cert_label)
 			ctx_log(ctx, 1, " ");
-		if (cert_label != NULL)
+		if (cert_label)
 			ctx_log(ctx, 1, "label=%s", cert_label);
 		ctx_log(ctx, 1, "\n");
 	}
 
+	matched_slots = (PKCS11_SLOT **)calloc(ctx->slot_count,
+		sizeof(PKCS11_SLOT *));
+	if (!matched_slots) {
+		ctx_log(ctx, 0, "Could not allocate memory for matched slots\n");
+		goto error;
+	}
+
 	for (n = 0; n < ctx->slot_count; n++) {
 		slot = ctx->slot_list + n;
 		flags[0] = '\0';
@@ -463,14 +469,15 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 			slot_nr == (int)PKCS11_get_slotid_from_slot(slot)) {
 			found_slot = slot;
 		}
+
 		if (match_tok && slot->token &&
-				(match_tok->label == NULL ||
+				(!match_tok->label ||
 					!strcmp(match_tok->label, slot->token->label)) &&
-				(match_tok->manufacturer == NULL ||
+				(!match_tok->manufacturer ||
 					!strcmp(match_tok->manufacturer, slot->token->manufacturer)) &&
-				(match_tok->serialnr == NULL ||
+				(!match_tok->serialnr ||
 					!strcmp(match_tok->serialnr, slot->token->serialnr)) &&
-				(match_tok->model == NULL ||
+				(!match_tok->model ||
 					!strcmp(match_tok->model, slot->token->model))) {
 			found_slot = slot;
 		}
@@ -483,8 +490,126 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 				slot->token->label : "no label");
 		}
 		ctx_log(ctx, 1, "\n");
+
+		if (found_slot && found_slot->token && !found_slot->token->initialized)
+			ctx_log(ctx, 0, "Found uninitialized token\n");
+
+		/* Ignore slots without tokens or with uninitialized token */
+		if (found_slot && found_slot->token && found_slot->token->initialized) {
+			matched_slots[matched_count] = found_slot;
+			matched_count++;
+		}
+		found_slot = NULL;
 	}
 
+	if (matched_count == 0) {
+		if (match_tok) {
+			ctx_log(ctx, 0, "Specified object not found\n");
+			goto error;
+		}
+
+		/* If the legacy slot ID format was used */
+		if (slot_nr != -1) {
+			ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
+			goto error;
+		} else {
+			found_slot = PKCS11_find_token(ctx->pkcs11_ctx,
+								ctx->slot_list, ctx->slot_count);
+			/* Ignore if the the token is not initialized */
+			if (found_slot && found_slot->token &&
+					found_slot->token->initialized) {
+				matched_slots[matched_count] = found_slot;
+				matched_count++;
+			} else {
+				ctx_log(ctx, 0, "No tokens found\n");
+				goto error;
+			}
+		}
+	}
+
+	for (n = 0; n < matched_count; n++) {
+		slot = matched_slots[n];
+		tok = slot->token;
+		if (!tok) {
+			ctx_log(ctx, 0, "Empty token found\n");
+			break;
+		}
+
+		ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
+		ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
+
+		/* In several tokens certificates are marked as private */
+		if (login) {
+			/* Only try to login if login is required */
+			if (tok->loginRequired) {
+				/* Only try to login if a single slot matched to avoiding trying
+				 * the PIN against all matching slots */
+				if (matched_count == 1) {
+					if (!ctx_login(ctx, slot, tok,
+							ctx->ui_method, ctx->callback_data)) {
+						ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
+						goto error;
+					}
+				} else {
+					ctx_log(ctx, 0, "Multiple matching slots (%lu); will not try to"
+						" login\n", matched_count);
+					for (m = 0; m < matched_count; m++){
+						slot = matched_slots[m];
+						ctx_log(ctx, 0, "[%u] %s: %s\n", m + 1,
+								slot->description? slot->description:
+								"(no description)",
+								(slot->token && slot->token->label)?
+								slot->token->label: "no label");
+					}
+					goto error;
+				}
+			}
+		}
+
+		if (PKCS11_enumerate_certs(tok, &certs, &cert_count)) {
+			ctx_log(ctx, 0, "Unable to enumerate certificates\n");
+			continue;
+		}
+
+		ctx_log(ctx, 1, "Found %u cert%s:\n", cert_count,
+				(cert_count <= 1) ? "" : "s");
+		if ((s_slot_cert_id && *s_slot_cert_id) &&
+				(cert_id_len != 0 || cert_label)) {
+			for (m = 0; m < cert_count; m++) {
+				PKCS11_CERT *k = certs + m;
+
+				if (cert_label && k->label && strcmp(k->label, cert_label) == 0)
+					selected_cert = k;
+				if (cert_id_len != 0 && k->id_len == cert_id_len &&
+						memcmp(k->id, cert_id, cert_id_len) == 0)
+					selected_cert = k;
+			}
+		} else {
+			for (m = 0; m < cert_count; m++) {
+				PKCS11_CERT *k = certs + m;
+				if (k->id && *(k->id)) {
+					selected_cert = k; /* Use the first certificate with nonempty id */
+					break;
+				}
+			}
+			if (!selected_cert)
+				selected_cert = certs; /* Use the first certificate */
+		}
+
+		if (selected_cert) {
+			break;
+		}
+	}
+
+	if (selected_cert) {
+		x509 = X509_dup(selected_cert->x509);
+	} else {
+		if (login) /* Only print the error on the second attempt */
+			ctx_log(ctx, 0, "Certificate not found.\n");
+		x509 = NULL;
+	}
+error:
+	/* Free the searched token data */
 	if (match_tok) {
 		OPENSSL_free(match_tok->model);
 		OPENSSL_free(match_tok->manufacturer);
@@ -492,77 +617,11 @@ static X509 *ctx_load_cert(ENGINE_CTX *ctx, const char *s_slot_cert_id,
 		OPENSSL_free(match_tok->label);
 		OPENSSL_free(match_tok);
 	}
-	if (found_slot) {
-		slot = found_slot;
-	} else if (match_tok) {
-		ctx_log(ctx, 0, "Specified object not found\n");
-		return NULL;
-	} else if (slot_nr == -1) {
-		if (!(slot = PKCS11_find_token(ctx->pkcs11_ctx,
-				ctx->slot_list, ctx->slot_count))) {
-			ctx_log(ctx, 0, "No tokens found\n");
-			return NULL;
-		}
-	} else {
-		ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
-		return NULL;
-	}
-	tok = slot->token;
 
-	if (tok == NULL) {
-		ctx_log(ctx, 0, "Empty token found\n");
-		return NULL;
-	}
-
-	ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
-	ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
-
-	/* In several tokens certificates are marked as private */
-	if (login && !ctx_login(ctx, slot, tok,
-			ctx->ui_method, ctx->callback_data)) {
-		ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
-		return NULL;
-	}
-
-	if (PKCS11_enumerate_certs(tok, &certs, &cert_count)) {
-		ctx_log(ctx, 0, "Unable to enumerate certificates\n");
-		return NULL;
-	}
-
-	ctx_log(ctx, 1, "Found %u cert%s:\n", cert_count,
-		(cert_count <= 1) ? "" : "s");
-	if ((s_slot_cert_id && *s_slot_cert_id) &&
-			(cert_id_len != 0 || cert_label != NULL)) {
-		for (n = 0; n < cert_count; n++) {
-			PKCS11_CERT *k = certs + n;
-
-			if (cert_label != NULL && strcmp(k->label, cert_label) == 0)
-				selected_cert = k;
-			if (cert_id_len != 0 && k->id_len == cert_id_len &&
-					memcmp(k->id, cert_id, cert_id_len) == 0)
-				selected_cert = k;
-		}
-	} else { 
-		for (n = 0; n < cert_count; n++) {
-			PKCS11_CERT *k = certs + n;
-			if(k->id && *(k->id)) {
-				selected_cert = k; /* Use the first certificate with nonempty id */
-				break;
-			}
-		}
-		if(!selected_cert)
-			selected_cert = certs; /* Use the first certificate */
-	}
-
-	if (selected_cert != NULL) {
-		x509 = X509_dup(selected_cert->x509);
-	} else {
-		if (login) /* Only print the error on the second attempt */
-			ctx_log(ctx, 0, "Certificate not found.\n");
-		x509 = NULL;
-	}
-	if (cert_label != NULL)
+	if (cert_label)
 		OPENSSL_free(cert_label);
+	if (matched_slots)
+		free(matched_slots);
 	return x509;
 }
 
@@ -573,22 +632,22 @@ static int ctx_ctrl_load_cert(ENGINE_CTX *ctx, void *p)
 		X509 *cert;
 	} *parms = p;
 
-	if (parms == NULL) {
+	if (!parms) {
 		ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ERR_R_PASSED_NULL_PARAMETER);
 		return 0;
 	}
-	if (parms->cert != NULL) {
+	if (parms->cert) {
 		ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ENG_R_INVALID_PARAMETER);
 		return 0;
 	}
 	ERR_clear_error();
 	if (!ctx->force_login)
 		parms->cert = ctx_load_cert(ctx, parms->s_slot_cert_id, 0);
-	if (parms->cert == NULL) { /* Try again with login */
+	if (!parms->cert) { /* Try again with login */
 		ERR_clear_error();
 		parms->cert = ctx_load_cert(ctx, parms->s_slot_cert_id, 1);
 	}
-	if (parms->cert == NULL) {
+	if (!parms->cert) {
 		if (!ERR_peek_last_error())
 			ENGerr(ENG_F_CTX_CTRL_LOAD_CERT, ENG_R_OBJECT_NOT_FOUND);
 		return 0;
@@ -605,7 +664,7 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 		const int isPrivate, const int login)
 {
 	PKCS11_SLOT *slot;
-	PKCS11_SLOT *found_slot = NULL;
+	PKCS11_SLOT *found_slot = NULL, **matched_slots = NULL;
 	PKCS11_TOKEN *tok, *match_tok = NULL;
 	PKCS11_KEY *keys, *selected_key = NULL;
 	EVP_PKEY *pk = NULL;
@@ -617,6 +676,7 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 	char tmp_pin[MAX_PIN_LENGTH+1];
 	size_t tmp_pin_len = MAX_PIN_LENGTH;
 	char flags[64];
+	size_t matched_count = 0;
 
 	if (ctx_init_libp11(ctx)) /* Delayed libp11 initialization */
 		goto error;
@@ -637,11 +697,13 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 				goto error;
 			}
 			if (tmp_pin_len > 0 && tmp_pin[0] != 0) {
-				if (!login)
+				/* If the searched key is public, try without login once even
+				 * when the PIN is provided */
+				if (!login && isPrivate)
 					goto error; /* Process on second attempt */
 				ctx_destroy_pin(ctx);
 				ctx->pin = OPENSSL_malloc(MAX_PIN_LENGTH+1);
-				if (ctx->pin != NULL) {
+				if (ctx->pin) {
 					memset(ctx->pin, 0, MAX_PIN_LENGTH+1);
 					memcpy(ctx->pin, tmp_pin, tmp_pin_len);
 					ctx->pin_length = tmp_pin_len;
@@ -666,13 +728,20 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			ctx_log(ctx, 1, "id=");
 			dump_hex(ctx, 1, key_id, key_id_len);
 		}
-		if (key_id_len != 0 && key_label != NULL)
+		if (key_id_len != 0 && key_label)
 			ctx_log(ctx, 1, " ");
-		if (key_label != NULL)
+		if (key_label)
 			ctx_log(ctx, 1, "label=%s", key_label);
 		ctx_log(ctx, 1, "\n");
 	}
 
+	matched_slots = (PKCS11_SLOT **)calloc(ctx->slot_count,
+		sizeof(PKCS11_SLOT *));
+	if (!matched_slots) {
+		ctx_log(ctx, 0, "Could not allocate memory for matched slots\n");
+		goto error;
+	}
+
 	for (n = 0; n < ctx->slot_count; n++) {
 		slot = ctx->slot_list + n;
 		flags[0] = '\0';
@@ -696,14 +765,15 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			slot_nr == (int)PKCS11_get_slotid_from_slot(slot)) {
 			found_slot = slot;
 		}
+
 		if (match_tok && slot->token &&
-				(match_tok->label == NULL ||
+				(!match_tok->label ||
 					!strcmp(match_tok->label, slot->token->label)) &&
-				(match_tok->manufacturer == NULL ||
+				(!match_tok->manufacturer ||
 					!strcmp(match_tok->manufacturer, slot->token->manufacturer)) &&
-				(match_tok->serialnr == NULL ||
+				(!match_tok->serialnr ||
 					!strcmp(match_tok->serialnr, slot->token->serialnr)) &&
-				(match_tok->model == NULL ||
+				(!match_tok->model ||
 					!strcmp(match_tok->model, slot->token->model))) {
 			found_slot = slot;
 		}
@@ -716,99 +786,130 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 				slot->token->label : "no label");
 		}
 		ctx_log(ctx, 1, "\n");
+
+		if (found_slot && found_slot->token && !found_slot->token->initialized)
+			ctx_log(ctx, 0, "Found uninitialized token\n");
+
+		/* Ignore slots without tokens or with uninitialized token */
+		if (found_slot && found_slot->token && found_slot->token->initialized) {
+			matched_slots[matched_count] = found_slot;
+			matched_count++;
+		}
+		found_slot = NULL;
 	}
 
-	if (match_tok) {
-		OPENSSL_free(match_tok->model);
-		OPENSSL_free(match_tok->manufacturer);
-		OPENSSL_free(match_tok->serialnr);
-		OPENSSL_free(match_tok->label);
-		OPENSSL_free(match_tok);
-	}
-	if (found_slot) {
-		slot = found_slot;
-	} else if (match_tok) {
-		ctx_log(ctx, 0, "Specified object not found\n");
-		goto error;
-	} else if (slot_nr == -1) {
-		if (!(slot = PKCS11_find_token(ctx->pkcs11_ctx,
-				ctx->slot_list, ctx->slot_count))) {
-			ctx_log(ctx, 0, "No tokens found\n");
+	if (matched_count == 0) {
+		if (match_tok) {
+			ctx_log(ctx, 0, "Specified object not found\n");
 			goto error;
 		}
-	} else {
-		ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
-		goto error;
-	}
-	tok = slot->token;
 
-	if (tok == NULL) {
-		ctx_log(ctx, 0, "Found empty token\n");
-		goto error;
-	}
-	/* The following check is non-critical to ensure interoperability
-	 * with some other (which ones?) PKCS#11 libraries */
-	if (!tok->initialized)
-		ctx_log(ctx, 0, "Found uninitialized token\n");
-	if (isPrivate && !tok->userPinSet && !tok->readOnly) {
-		ctx_log(ctx, 0, "Found slot without user PIN\n");
-		goto error;
-	}
-
-	ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
-	ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
-
-	/* Both private and public keys can have the CKA_PRIVATE attribute
-	 * set and thus require login (even to retrieve attributes!) */
-	if (login && !ctx_login(ctx, slot, tok, ui_method, callback_data)) {
-		ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
-		goto error;
-	}
-
-	if (isPrivate) {
-		/* Make sure there is at least one private key on the token */
-		if (PKCS11_enumerate_keys(tok, &keys, &key_count)) {
-			ctx_log(ctx, 0, "Unable to enumerate private keys\n");
-			goto error;
-		}
-	} else {
-		/* Make sure there is at least one public key on the token */
-		if (PKCS11_enumerate_public_keys(tok, &keys, &key_count)) {
-			ctx_log(ctx, 0, "Unable to enumerate public keys\n");
+		/* If the legacy slot ID format was used */
+		if (slot_nr != -1) {
+			ctx_log(ctx, 0, "Invalid slot number: %d\n", slot_nr);
 			goto error;
+		} else {
+			found_slot = PKCS11_find_token(ctx->pkcs11_ctx,
+								ctx->slot_list, ctx->slot_count);
+			/* Ignore if the the token is not initialized */
+			if (found_slot && found_slot->token &&
+					found_slot->token->initialized) {
+				matched_slots[matched_count] = found_slot;
+				matched_count++;
+			} else {
+				ctx_log(ctx, 0, "No tokens found\n");
+				goto error;
+			}
 		}
 	}
-	if (key_count == 0) {
-		if (login) /* Only print the error on the second attempt */
-			ctx_log(ctx, 0, "No %s keys found.\n",
-				(char *)(isPrivate ? "private" : "public"));
-		goto error;
-	}
-	ctx_log(ctx, 1, "Found %u %s key%s:\n", key_count,
-		(char *)(isPrivate ? "private" : "public"),
-		(key_count == 1) ? "" : "s");
 
-	if (s_slot_key_id && *s_slot_key_id &&
-			(key_id_len != 0 || key_label != NULL)) {
-		for (n = 0; n < key_count; n++) {
-			PKCS11_KEY *k = keys + n;
-
-			ctx_log(ctx, 1, "  %2u %c%c id=", n + 1,
-				k->isPrivate ? 'P' : ' ',
-				k->needLogin ? 'L' : ' ');
-			dump_hex(ctx, 1, k->id, k->id_len);
-			ctx_log(ctx, 1, " label=%s\n", k->label);
-			if (key_label != NULL && strcmp(k->label, key_label) == 0)
-				selected_key = k;
-			if (key_id_len != 0 && k->id_len == key_id_len
-					&& memcmp(k->id, key_id, key_id_len) == 0)
-				selected_key = k;
+	for (n = 0; n < matched_count; n++) {
+		slot = matched_slots[n];
+		tok = slot->token;
+		if (!tok) {
+			ctx_log(ctx, 0, "Found empty token\n");
+			break;
+		}
+
+		ctx_log(ctx, 1, "Found slot:  %s\n", slot->description);
+		ctx_log(ctx, 1, "Found token: %s\n", slot->token->label);
+
+		/* Both private and public keys can have the CKA_PRIVATE attribute
+		 * set and thus require login (even to retrieve attributes!) */
+		if (login) {
+			/* Try to login only if login is required */
+			if (tok->loginRequired) {
+				/* Try to login only if a single slot matched to avoiding trying
+				 * the PIN against all matching slots */
+				if (matched_count == 1) {
+					if (!ctx_login(ctx, slot, tok, ui_method, callback_data)) {
+						ctx_log(ctx, 0, "Login to token failed, returning NULL...\n");
+						goto error;
+					}
+				} else {
+					ctx_log(ctx, 0, "Multiple matching slots (%lu); will not try to"
+						" login\n", matched_count);
+					for (m = 0; m < matched_count; m++){
+						slot = matched_slots[m];
+						ctx_log(ctx, 1, "[%u] %s: %s\n", m + 1,
+								slot->description? slot->description:
+								"(no description)",
+								(slot->token && slot->token->label)?
+								slot->token->label: "no label");
+					}
+					goto error;
+				}
+			}
+		}
+
+		if (isPrivate) {
+			/* Make sure there is at least one private key on the token */
+			if (PKCS11_enumerate_keys(tok, &keys, &key_count)) {
+				ctx_log(ctx, 0, "Unable to enumerate private keys\n");
+				continue;
+			}
+		} else {
+			/* Make sure there is at least one public key on the token */
+			if (PKCS11_enumerate_public_keys(tok, &keys, &key_count)) {
+				ctx_log(ctx, 0, "Unable to enumerate public keys\n");
+				continue;
+			}
+		}
+		if (key_count == 0) {
+			if (login) /* Only print the error on the second attempt */
+				ctx_log(ctx, 0, "No %s keys found.\n",
+						(char *)(isPrivate ? "private" : "public"));
+			continue;
+		}
+		ctx_log(ctx, 1, "Found %u %s key%s:\n", key_count,
+				(char *)(isPrivate ? "private" : "public"),
+				(key_count == 1) ? "" : "s");
+
+		if (s_slot_key_id && *s_slot_key_id && (key_id_len != 0 || key_label)) {
+			for (m = 0; m < key_count; m++) {
+				PKCS11_KEY *k = keys + m;
+
+				ctx_log(ctx, 1, "  %2u %c%c id=", m + 1,
+						k->isPrivate ? 'P' : ' ',
+						k->needLogin ? 'L' : ' ');
+				dump_hex(ctx, 1, k->id, k->id_len);
+				ctx_log(ctx, 1, " label=%s\n", k->label ? k->label : "(null)");
+				if (key_label && k->label && strcmp(k->label, key_label) == 0)
+					selected_key = k;
+				if (key_id_len != 0 && k->id_len == key_id_len
+						&& memcmp(k->id, key_id, key_id_len) == 0)
+					selected_key = k;
+			}
+		} else {
+			selected_key = keys; /* Use the first key */
+		}
+
+		if (selected_key) {
+			break;
 		}
-	} else {
-		selected_key = keys; /* Use the first key */
 	}
 
-	if (selected_key != NULL) {
+	if (selected_key) {
 		pk = isPrivate ?
 			PKCS11_get_private_key(selected_key) :
 			PKCS11_get_public_key(selected_key);
@@ -817,9 +918,20 @@ static EVP_PKEY *ctx_load_key(ENGINE_CTX *ctx, const char *s_slot_key_id,
 			ctx_log(ctx, 0, "Key not found.\n");
 		pk = NULL;
 	}
+
 error:
-	if (key_label != NULL)
+	/* Free the searched token data */
+	if (match_tok) {
+		OPENSSL_free(match_tok->model);
+		OPENSSL_free(match_tok->manufacturer);
+		OPENSSL_free(match_tok->serialnr);
+		OPENSSL_free(match_tok->label);
+		OPENSSL_free(match_tok);
+	}
+	if (key_label)
 		OPENSSL_free(key_label);
+	if (matched_slots)
+		free(matched_slots);
 	return pk;
 }
 
@@ -831,11 +943,11 @@ EVP_PKEY *ctx_load_pubkey(ENGINE_CTX *ctx, const char *s_key_id,
 	ERR_clear_error();
 	if (!ctx->force_login)
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 0, 0);
-	if (pk == NULL) { /* Try again with login */
+	if (!pk) { /* Try again with login */
 		ERR_clear_error();
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 0, 1);
 	}
-	if (pk == NULL) {
+	if (!pk) {
 		ctx_log(ctx, 0, "PKCS11_load_public_key returned NULL\n");
 		if (!ERR_peek_last_error())
 			ENGerr(ENG_F_CTX_LOAD_PUBKEY, ENG_R_OBJECT_NOT_FOUND);
@@ -852,11 +964,11 @@ EVP_PKEY *ctx_load_privkey(ENGINE_CTX *ctx, const char *s_key_id,
 	ERR_clear_error();
 	if (!ctx->force_login)
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 1, 0);
-	if (pk == NULL) { /* Try again with login */
+	if (!pk) { /* Try again with login */
 		ERR_clear_error();
 		pk = ctx_load_key(ctx, s_key_id, ui_method, callback_data, 1, 1);
 	}
-	if (pk == NULL) {
+	if (!pk) {
 		ctx_log(ctx, 0, "PKCS11_get_private_key returned NULL\n");
 		if (!ERR_peek_last_error())
 			ENGerr(ENG_F_CTX_LOAD_PRIVKEY, ENG_R_OBJECT_NOT_FOUND);
@@ -892,7 +1004,7 @@ static int ctx_ctrl_set_module(ENGINE_CTX *ctx, const char *modulename)
 static int ctx_ctrl_set_pin(ENGINE_CTX *ctx, const char *pin)
 {
 	/* Pre-condition check */
-	if (pin == NULL) {
+	if (!pin) {
 		ENGerr(ENG_F_CTX_CTRL_SET_PIN, ERR_R_PASSED_NULL_PARAMETER);
 		errno = EINVAL;
 		return 0;
@@ -902,7 +1014,7 @@ static int ctx_ctrl_set_pin(ENGINE_CTX *ctx, const char *pin)
 	 * shall be returned and errno shall be set. */
 	ctx_destroy_pin(ctx);
 	ctx->pin = OPENSSL_strdup(pin);
-	if (ctx->pin == NULL) {
+	if (!ctx->pin) {
 		ENGerr(ENG_F_CTX_CTRL_SET_PIN, ERR_R_MALLOC_FAILURE);
 		errno = ENOMEM;
 		return 0;
@@ -933,7 +1045,7 @@ static int ctx_ctrl_set_init_args(ENGINE_CTX *ctx, const char *init_args_orig)
 static int ctx_ctrl_set_user_interface(ENGINE_CTX *ctx, UI_METHOD *ui_method)
 {
 	ctx->ui_method = ui_method;
-	if (ctx->pkcs11_ctx != NULL) /* libp11 is already initialized */
+	if (ctx->pkcs11_ctx) /* libp11 is already initialized */
 		PKCS11_set_ui_method(ctx->pkcs11_ctx,
 			ctx->ui_method, ctx->callback_data);
 	return 1;
@@ -942,7 +1054,7 @@ static int ctx_ctrl_set_user_interface(ENGINE_CTX *ctx, UI_METHOD *ui_method)
 static int ctx_ctrl_set_callback_data(ENGINE_CTX *ctx, void *callback_data)
 {
 	ctx->callback_data = callback_data;
-	if (ctx->pkcs11_ctx != NULL) /* libp11 is already initialized */
+	if (ctx->pkcs11_ctx) /* libp11 is already initialized */
 		PKCS11_set_ui_method(ctx->pkcs11_ctx,
 			ctx->ui_method, ctx->callback_data);
 	return 1;
